from .ShiftedReLU import ShiftedReLU
from .PLogish import PLogish
from .ParametricLogish import ParametricLogish
from .ExpExpish import ExpExpish
from .GeneralizedSwish import GeneralizedSwish
from .TBSReLU import TBSReLU
from .NoisyReLU import NoisyReLU
from .ExponentialDLReLU import ExponentialDLReLU
from .SReLU import SReLU
from .TangentSigmoidReLU import TangentSigmoidReLU
from .Phish import Phish
from .WideMeLU import WideMeLU
from .SelfArctan import SelfArctan
from .LogSigmoid import LogSigmoid
from .SlopedReLU import SlopedReLU
from .SmallGaLU import SmallGaLU
from .MinSin import MinSin
from .LaLU import LaLU
from .MexicanReLU import MexicanReLU
from .APLU import APLU
from .ERF import ERF
from .TangentBipolarSigmoidReLU import TangentBipolarSigmoidReLU
from .BaseDLReLU import BaseDLReLU
from .ShiLU import ShiLU
from .Logish import Logish
from .TripleStateSwish import TripleStateSwish
from .ExponentialSwish import ExponentialSwish
from .TeLU import TeLU
from .Elliot import Elliot
from .MeLU import MeLU
from .GaussianReLU import GaussianReLU
from .ActivationFunction import ActivationFunction
from .RReLU import RReLU
from .learnable.AOAF import AOAF
from .Suish import Suish
from .SoftsignRReLU import SoftsignRReLU
from .Gish import Gish
from .NReLU import NReLU
from .LogLogish import LogLogish
from .SGELU import SGELU
from .GaLU import GaLU
from .TSReLU import TSReLU
from .SineReLU import SineReLU
from .DLReLU import DLReLU
from .CaLU import CaLU
from .RandomizedSlopedReLU import RandomizedSlopedReLU
from .GCU import GCU
from .SigmoidDerivative import SigmoidDerivative
from .Smish import Smish
from .AdaptiveActivationFunction import AdaptiveActivationFunction
from .learnable import *



import json


def settings(key: str, _cache: dict = {}):
    if not _cache:
        with open("settings.json", "r") as f:
            _cache.update(json.load(f))
    return _cache.get(key)


def get_settings(*args, **kwargs):
    return settings(*args, **kwargs)



import torch
from .ActivationFunction import ActivationFunction


class Elliot(ActivationFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return 0.5 + torch.div(0.5 * x, 1 + torch.abs(x))



from pathlib import Path


class MakeNamespace:
    def trait(self, name: str):
        class_name = name.split(".")[-1]
        file = name.replace(".", "/") + ".py"
        path = Path(file)
        parent = path.parent
        parent.mkdir(parents=True, exist_ok=True)
        path.touch()

        # copy the tempalte path
        current_path = Path(__file__).parent.parent.resolve()
        template_path = current_path / "templates" / "trait.txt"
        content = template_path.read_text()
        content = content.replace("{{CLASS}}", class_name)
        path.write_text(content)



import torch
from typing import List


class ChannelBased:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initialized = False
        self.num_channels = None

    def initialize(
        self, x: torch.Tensor, attrs: List[str] | str, values: List[float] | float = []
    ):
        if getattr(self, "_initialized", False):
            return

        if not isinstance(values, list):
            values = [values]

        if not isinstance(attrs, list):
            attrs = [attrs]

        self.num_channels = x.shape[1]
        for index, attr in enumerate(attrs):
            if index < len(values) and values[index] is not None:
                default_value = float(values[index])
            else:
                default_value = 1.0
            param = torch.nn.Parameter(torch.full((self.num_channels,), default_value))
            setattr(self, attr, param)
        self._initialized = True

    def parameter_shape(self, x: torch.Tensor) -> tuple | None:
        if hasattr(self, "num_channels"):
            return (1, self.num_channels) + (1,) * (x.ndim - 2)
        return None



import torch


class HasLogits:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def logits(self, x: torch.Tensor) -> torch.Tensor:
        return self.network(x)

    def logits_(self, x: torch.Tensor) -> torch.Tensor:
        return self.network(x)



from torch import Tensor, no_grad
from torch.utils.data.dataloader import DataLoader
from ...device import get_device


class ReportsAccuracy:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def accuracy_tensors(self, x: Tensor, y: Tensor) -> float:
        predictions = self.predict_(x)
        return (predictions == y).float().mean().item()

    def accuracy_dataloader(self, dataloader: DataLoader):
        correct, total = 0, 0
        for x, y in dataloader:
            x, y = x.to(get_device()), y.to(get_device())
            predictions = self.predict_(x)
            correct += (predictions == y).sum().item()
            total += y.numel()
        return correct / total if total > 0 else 0.0

    def accuracy(self, x: Tensor | DataLoader, y: Tensor | None = None) -> float:
        self.network.eval()
        if isinstance(x, DataLoader):
            return self.accuracy_dataloader(x)
        return self.accuracy_tensors(x, y)



import torch


class ERF(torch.nn.Module):
    def __init__(self, alpha=1.0):
        super().__init__()
        self.alpha = torch.nn.Parameter(torch.tensor(alpha))

    def forward(self, x: torch.Tensor):
        return x * torch.erf(self.alpha * x)
